"""
Tests for git execution module

NOTE: These tests may modify the global git config, and have been marked as
destructive as a result. If no values are set for user.name or user.email in
the user's global .gitconfig, then these tests will set one.
"""

import logging
import os
import pathlib
import re
import shutil
import subprocess
import tarfile
import tempfile
from contextlib import closing

import pytest

import salt.utils.data
import salt.utils.files
import salt.utils.platform
from salt.utils.versions import LooseVersion
from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


def _git_version():
    git = shutil.which("git")
    if not git:
        log.debug("Git not installed")
        return False
    ret = subprocess.run(
        ["git", "--version"],
        stdout=subprocess.PIPE,
        check=False,
        shell=False,
        universal_newlines=True,
    )
    # On macOS, the git version is displayed in a different format
    #  git version 2.21.1 (Apple Git-122.3)
    # As opposed to:
    #  git version 2.21.1
    version_str = ret.stdout.strip().split("(")[0].strip().split()[-1]
    git_version = LooseVersion(version_str)
    log.debug("Detected git version: %s", git_version)
    return git_version


def _worktrees_supported():
    """
    Check if the git version is 2.5.0 or later
    """
    try:
        return _git_version() >= LooseVersion("2.5.0")
    except AttributeError:
        return False


@pytest.mark.windows_whitelisted
@pytest.mark.skip_if_binaries_missing("git")
class GitModuleTest(ModuleCase):
    def setUp(self):
        super().setUp()
        self.repo = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.repo, ignore_errors=True)
        self.addCleanup(delattr, self, "repo")
        self.files = ("foo", "bar", "baz", "питон")
        self.addCleanup(delattr, self, "files")
        self.dirs = ("", "qux")
        self.addCleanup(delattr, self, "dirs")
        self.branches = ("master", "iamanewbranch")
        self.addCleanup(delattr, self, "branches")
        self.tags = ("git_testing",)
        self.addCleanup(delattr, self, "tags")
        for dirname in self.dirs:
            dir_path = pathlib.Path(self.repo) / dirname
            dir_path.mkdir(parents=True, exist_ok=True)
            for filename in self.files:
                with salt.utils.files.fopen(str(dir_path / filename), "wb") as fp_:
                    fp_.write(
                        "This is a test file named {}.".format(filename).encode("utf-8")
                    )
        # Navigate to the root of the repo to init, stage, and commit
        with pytest.helpers.change_cwd(self.repo):
            # Initialize a new git repository
            subprocess.check_call(["git", "init", "--quiet", self.repo])

            # Set user.name and user.email config attributes if not present
            for key, value in (
                ("user.name", "Jenkins"),
                ("user.email", "qa@saltstack.com"),
            ):
                # Check if key is missing
                keycheck = subprocess.Popen(
                    ["git", "config", "--get", "--global", key],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                if keycheck.wait() != 0:
                    # Set the key if it is not present
                    subprocess.check_call(["git", "config", "--global", key, value])

            subprocess.check_call(["git", "add", "."])
            subprocess.check_call(
                ["git", "commit", "--quiet", "--message", "Initial commit"]
            )
            # Add a tag
            subprocess.check_call(["git", "tag", "-a", self.tags[0], "-m", "Add tag"])
            # Checkout a second branch
            subprocess.check_call(
                ["git", "checkout", "--quiet", "-b", self.branches[1]]
            )
            # Add a line to the file
            with salt.utils.files.fopen(self.files[0], "a") as fp_:
                fp_.write(salt.utils.stringutils.to_str("Added a line\n"))
            # Commit the updated file
            subprocess.check_call(
                [
                    "git",
                    "commit",
                    "--quiet",
                    "--message",
                    "Added a line to " + self.files[0],
                    self.files[0],
                ]
            )
            # Switch back to master
            subprocess.check_call(["git", "checkout", "--quiet", "master"])

    def run_function(self, *args, **kwargs):  # pylint: disable=arguments-differ
        """
        Ensure that results are decoded

        TODO: maybe move this behavior to ModuleCase itself?
        """
        return salt.utils.data.decode(super().run_function(*args, **kwargs))

    @pytest.mark.slow_test
    def test_add_dir(self):
        """
        Test git.add with a directory
        """
        newdir = "quux"
        # Change to the repo dir
        newdir_path = pathlib.Path(self.repo) / newdir
        newdir_path.mkdir(parents=True, exist_ok=True)
        files = [str(newdir_path / x) for x in self.files]
        files_relpath = [os.path.join(newdir, x) for x in self.files]
        for path in files:
            with salt.utils.files.fopen(path, "wb") as fp_:
                fp_.write(
                    "This is a test file with relative path {}.\n".format(path).encode(
                        "utf-8"
                    )
                )
        ret = self.run_function("git.add", [self.repo, newdir])
        res = "\n".join(sorted("add '{}'".format(x) for x in files_relpath))
        if salt.utils.platform.is_windows():
            res = res.replace("\\", "/")
        self.assertEqual(ret, res)

    @pytest.mark.slow_test
    def test_add_file(self):
        """
        Test git.add with a file
        """
        filename = "quux"
        file_path = os.path.join(self.repo, filename)
        with salt.utils.files.fopen(file_path, "w") as fp_:
            fp_.write(
                salt.utils.stringutils.to_str(
                    "This is a test file named {}.\n".format(filename)
                )
            )
        ret = self.run_function("git.add", [self.repo, filename])
        self.assertEqual(ret, "add '{}'".format(filename))

    @pytest.mark.slow_test
    def test_archive(self):
        """
        Test git.archive
        """
        tar_archive = os.path.join(RUNTIME_VARS.TMP, "test_archive.tar.gz")
        try:
            self.assertTrue(
                self.run_function(
                    "git.archive", [self.repo, tar_archive], prefix="foo/"
                )
            )
            self.assertTrue(tarfile.is_tarfile(tar_archive))
            self.run_function("cmd.run", ["cp " + tar_archive + " /root/"])
            with closing(tarfile.open(tar_archive, "r")) as tar_obj:
                self.assertEqual(
                    sorted(salt.utils.data.decode(tar_obj.getnames())),
                    sorted(
                        [
                            "foo",
                            "foo/bar",
                            "foo/baz",
                            "foo/foo",
                            "foo/питон",
                            "foo/qux",
                            "foo/qux/bar",
                            "foo/qux/baz",
                            "foo/qux/foo",
                            "foo/qux/питон",
                        ]
                    ),
                )
        finally:
            try:
                os.unlink(tar_archive)
            except OSError:
                pass

    @pytest.mark.slow_test
    def test_archive_subdir(self):
        """
        Test git.archive on a subdir, giving only a partial copy of the repo in
        the resulting archive
        """
        tar_archive = os.path.join(RUNTIME_VARS.TMP, "test_archive.tar.gz")
        try:
            self.assertTrue(
                self.run_function(
                    "git.archive",
                    [os.path.join(self.repo, "qux"), tar_archive],
                    prefix="foo/",
                )
            )
            self.assertTrue(tarfile.is_tarfile(tar_archive))
            with closing(tarfile.open(tar_archive, "r")) as tar_obj:
                self.assertEqual(
                    sorted(salt.utils.data.decode(tar_obj.getnames())),
                    sorted(["foo", "foo/bar", "foo/baz", "foo/foo", "foo/питон"]),
                )
        finally:
            try:
                os.unlink(tar_archive)
            except OSError:
                pass

    @pytest.mark.slow_test
    def test_branch(self):
        """
        Test creating, renaming, and deleting a branch using git.branch
        """
        renamed_branch = "ihavebeenrenamed"
        self.assertTrue(self.run_function("git.branch", [self.repo, self.branches[1]]))
        self.assertTrue(
            self.run_function(
                "git.branch", [self.repo, renamed_branch], opts="-m " + self.branches[1]
            )
        )
        self.assertTrue(
            self.run_function("git.branch", [self.repo, renamed_branch], opts="-D")
        )

    @pytest.mark.slow_test
    def test_checkout(self):
        """
        Test checking out a new branch and then checking out master again
        """
        new_branch = "iamanothernewbranch"
        self.assertEqual(
            self.run_function(
                "git.checkout", [self.repo, "HEAD"], opts="-b " + new_branch
            ),
            "Switched to a new branch '" + new_branch + "'",
        )
        self.assertTrue(
            "Switched to branch 'master'"
            in self.run_function("git.checkout", [self.repo, "master"]),
        )

    @pytest.mark.slow_test
    def test_checkout_no_rev(self):
        """
        Test git.checkout without a rev, both with -b in opts and without
        """
        new_branch = "iamanothernewbranch"
        self.assertEqual(
            self.run_function(
                "git.checkout", [self.repo], rev=None, opts="-b " + new_branch
            ),
            "Switched to a new branch '" + new_branch + "'",
        )
        self.assertTrue(
            "'rev' argument is required unless -b or -B in opts"
            in self.run_function("git.checkout", [self.repo])
        )

    @pytest.mark.slow_test
    def test_clone(self):
        """
        Test cloning an existing repo
        """
        clone_parent_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.assertTrue(self.run_function("git.clone", [clone_parent_dir, self.repo]))
        # Cleanup after yourself
        shutil.rmtree(clone_parent_dir, True)

    @pytest.mark.slow_test
    def test_clone_with_alternate_name(self):
        """
        Test cloning an existing repo with an alternate name for the repo dir
        """
        clone_parent_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        clone_name = os.path.basename(self.repo)
        # Change to newly-created temp dir
        self.assertTrue(
            self.run_function(
                "git.clone", [clone_parent_dir, self.repo], name=clone_name
            )
        )
        # Cleanup after yourself
        shutil.rmtree(clone_parent_dir, True)

    @pytest.mark.slow_test
    def test_commit(self):
        """
        Test git.commit two ways:
            1) First using git.add, then git.commit
            2) Using git.commit with the 'filename' argument to skip staging
        """
        filename = "foo"
        commit_re_prefix = r"^\[master [0-9a-f]+\] "
        # Add a line
        with salt.utils.files.fopen(os.path.join(self.repo, filename), "a") as fp_:
            fp_.write("Added a line\n")
        # Stage the file
        self.run_function("git.add", [self.repo, filename])
        # Commit the staged file
        commit_msg = "Add a line to " + filename
        ret = self.run_function("git.commit", [self.repo, commit_msg])
        # Make sure the expected line is in the output
        self.assertTrue(bool(re.search(commit_re_prefix + commit_msg, ret)))
        # Add another line
        with salt.utils.files.fopen(os.path.join(self.repo, filename), "a") as fp_:
            fp_.write("Added another line\n")
        # Commit the second file without staging
        commit_msg = "Add another line to " + filename
        ret = self.run_function(
            "git.commit", [self.repo, commit_msg], filename=filename
        )
        self.assertTrue(bool(re.search(commit_re_prefix + commit_msg, ret)))

    @pytest.mark.slow_test
    def test_config(self):
        """
        Test setting, getting, and unsetting config values

        WARNING: This test will modify and completely remove a config section
        'foo', both in the repo created in setUp() and in the user's global
        .gitconfig.
        """

        def _clear_config():
            cmds = (
                ["git", "config", "--remove-section", "foo"],
                ["git", "config", "--global", "--remove-section", "foo"],
            )
            for cmd in cmds:
                with salt.utils.files.fopen(os.devnull, "w") as devnull:
                    try:
                        subprocess.check_call(cmd, stderr=devnull)
                    except subprocess.CalledProcessError:
                        pass

        cfg_local = {"foo.single": ["foo"], "foo.multi": ["foo", "bar", "baz"]}
        cfg_global = {"foo.single": ["abc"], "foo.multi": ["abc", "def", "ghi"]}
        _clear_config()
        try:
            log.debug("Try to specify both single and multivar (should raise error)")
            self.assertTrue(
                "Only one of 'value' and 'multivar' is permitted"
                in self.run_function(
                    "git.config_set",
                    ["foo.single"],
                    value=cfg_local["foo.single"][0],
                    multivar=cfg_local["foo.multi"],
                    cwd=self.repo,
                )
            )
            log.debug("Try to set single local value without cwd (should raise error)")
            self.assertTrue(
                "'cwd' argument required unless global=True"
                in self.run_function(
                    "git.config_set",
                    ["foo.single"],
                    value=cfg_local["foo.single"][0],
                )
            )
            log.debug("Set single local value")
            self.assertEqual(
                self.run_function(
                    "git.config_set",
                    ["foo.single"],
                    value=cfg_local["foo.single"][0],
                    cwd=self.repo,
                ),
                cfg_local["foo.single"],
            )
            log.debug("Set single global value")
            self.assertEqual(
                self.run_function(
                    "git.config_set",
                    ["foo.single"],
                    value=cfg_global["foo.single"][0],
                    **{"global": True}
                ),
                cfg_global["foo.single"],
            )
            log.debug("Set local multivar")
            self.assertEqual(
                self.run_function(
                    "git.config_set",
                    ["foo.multi"],
                    multivar=cfg_local["foo.multi"],
                    cwd=self.repo,
                ),
                cfg_local["foo.multi"],
            )
            log.debug("Set global multivar")
            self.assertEqual(
                self.run_function(
                    "git.config_set",
                    ["foo.multi"],
                    multivar=cfg_global["foo.multi"],
                    **{"global": True}
                ),
                cfg_global["foo.multi"],
            )
            log.debug("Get single local value")
            self.assertEqual(
                self.run_function("git.config_get", ["foo.single"], cwd=self.repo),
                cfg_local["foo.single"][0],
            )
            log.debug("Get single value from local multivar")
            self.assertEqual(
                self.run_function("git.config_get", ["foo.multi"], cwd=self.repo),
                cfg_local["foo.multi"][-1],
            )
            log.debug("Get all values from multivar (includes globals)")
            self.assertEqual(
                self.run_function(
                    "git.config_get", ["foo.multi"], cwd=self.repo, **{"all": True}
                ),
                cfg_local["foo.multi"],
            )
            log.debug("Get single global value")
            self.assertEqual(
                self.run_function("git.config_get", ["foo.single"], **{"global": True}),
                cfg_global["foo.single"][0],
            )
            log.debug("Get single value from global multivar")
            self.assertEqual(
                self.run_function("git.config_get", ["foo.multi"], **{"global": True}),
                cfg_global["foo.multi"][-1],
            )
            log.debug("Get all values from global multivar")
            self.assertEqual(
                self.run_function(
                    "git.config_get", ["foo.multi"], **{"all": True, "global": True}
                ),
                cfg_global["foo.multi"],
            )
            log.debug("Get all local keys/values using regex")
            self.assertEqual(
                self.run_function(
                    "git.config_get_regexp", ["foo.(single|multi)"], cwd=self.repo
                ),
                cfg_local,
            )
            log.debug("Get all global keys/values using regex")
            self.assertEqual(
                self.run_function(
                    "git.config_get_regexp",
                    ["foo.(single|multi)"],
                    cwd=self.repo,
                    **{"global": True}
                ),
                cfg_global,
            )
            log.debug("Get just the local foo.multi values containing 'a'")
            self.assertEqual(
                self.run_function(
                    "git.config_get_regexp",
                    ["foo.multi"],
                    value_regex="a",
                    cwd=self.repo,
                ),
                {"foo.multi": [x for x in cfg_local["foo.multi"] if "a" in x]},
            )
            log.debug("Get just the global foo.multi values containing 'a'")
            self.assertEqual(
                self.run_function(
                    "git.config_get_regexp",
                    ["foo.multi"],
                    value_regex="a",
                    cwd=self.repo,
                    **{"global": True}
                ),
                {"foo.multi": [x for x in cfg_global["foo.multi"] if "a" in x]},
            )

            # TODO: More robust unset testing, try to trigger all the
            # exceptions raised.

            log.debug("Unset a single local value")
            self.assertTrue(
                self.run_function(
                    "git.config_unset",
                    ["foo.single"],
                    cwd=self.repo,
                )
            )
            log.debug("Unset an entire local multivar")
            self.assertTrue(
                self.run_function(
                    "git.config_unset", ["foo.multi"], cwd=self.repo, **{"all": True}
                )
            )
            log.debug("Unset a single global value")
            self.assertTrue(
                self.run_function(
                    "git.config_unset", ["foo.single"], **{"global": True}
                )
            )
            log.debug("Unset an entire local multivar")
            self.assertTrue(
                self.run_function(
                    "git.config_unset", ["foo.multi"], **{"all": True, "global": True}
                )
            )
        finally:
            _clear_config()

    @pytest.mark.slow_test
    def test_current_branch(self):
        """
        Test git.current_branch
        """
        self.assertEqual(self.run_function("git.current_branch", [self.repo]), "master")

    @pytest.mark.slow_test
    def test_describe(self):
        """
        Test git.describe
        """
        self.assertEqual(self.run_function("git.describe", [self.repo]), self.tags[0])

    # Test for git.fetch would be unreliable on Jenkins, skipping for now
    # The test should go into test_remotes when ready

    @pytest.mark.slow_test
    def test_init(self):
        """
        Use git.init to init a new repo
        """
        new_repo = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        # `tempfile.mkdtemp` gets the path to the Temp directory using
        # environment variables. As a result, folder names longer than 8
        # characters are shortened. For example "C:\Users\Administrators"
        # becomes "C:\Users\Admini~1". However, the "git.init" function returns
        # the full, unshortened name of the folder. Therefore you can't compare
        # the path returned by `tempfile.mkdtemp` and the results of `git.init`
        # exactly.
        if salt.utils.platform.is_windows():
            new_repo = new_repo.replace("\\", "/")

            # Get the name of the temp directory
            tmp_dir = os.path.basename(new_repo)

            # Get git output
            git_ret = self.run_function("git.init", [new_repo]).lower()

            self.assertIn("Initialized empty Git repository in".lower(), git_ret)
            self.assertIn(tmp_dir, git_ret)

        else:
            self.assertEqual(
                self.run_function("git.init", [new_repo]).lower(),
                "Initialized empty Git repository in {}/.git/".format(new_repo).lower(),
            )

        shutil.rmtree(new_repo)

    @pytest.mark.slow_test
    def test_list_branches(self):
        """
        Test git.list_branches
        """
        self.assertEqual(
            self.run_function("git.list_branches", [self.repo]), sorted(self.branches)
        )

    @pytest.mark.slow_test
    def test_list_tags(self):
        """
        Test git.list_tags
        """
        self.assertEqual(
            self.run_function("git.list_tags", [self.repo]), sorted(self.tags)
        )

    # Test for git.ls_remote will need to wait for now, while I think of how to
    # properly mock it.

    @pytest.mark.slow_test
    def test_merge(self):
        """
        Test git.merge

        # TODO: Test more than just a fast-forward merge
        """
        # Merge the second branch into the current branch
        ret = self.run_function("git.merge", [self.repo], rev=self.branches[1])
        # Merge should be a fast-forward
        self.assertTrue("Fast-forward" in ret.splitlines())

    @pytest.mark.slow_test
    def test_merge_base_and_tree(self):
        """
        Test git.merge_base, git.merge_tree and git.revision

        TODO: Test all of the arguments
        """
        # Get the SHA1 of current HEAD
        head_rev = self.run_function("git.revision", [self.repo], rev="HEAD")
        # Make sure revision is a 40-char string
        self.assertTrue(len(head_rev) == 40)
        # Get the second branch's SHA1
        second_rev = self.run_function(
            "git.revision", [self.repo], rev=self.branches[1], timeout=120
        )
        # Make sure revision is a 40-char string
        self.assertTrue(len(second_rev) == 40)
        # self.branches[1] should be just one commit ahead, so the merge base
        # for master and self.branches[1] should be the same as the current
        # HEAD.
        self.assertEqual(
            self.run_function(
                "git.merge_base", [self.repo], refs=",".join((head_rev, second_rev))
            ),
            head_rev,
        )
        # There should be no conflict here, so the return should be an empty
        # string.
        ret = self.run_function(
            "git.merge_tree", [self.repo, head_rev, second_rev]
        ).splitlines()
        self.assertTrue(len([x for x in ret if x.startswith("@@")]) == 1)

    # Test for git.pull would be unreliable on Jenkins, skipping for now

    # Test for git.push would be unreliable on Jenkins, skipping for now

    @pytest.mark.slow_test
    def test_rebase(self):
        """
        Test git.rebase
        """
        # Switch to the second branch
        self.assertNotIn(
            "ERROR",
            self.run_function("git.checkout", [self.repo], rev=self.branches[0]),
        )
        # Make a change to a different file than the one modified in setUp
        file_path = os.path.join(self.repo, self.files[1])
        with salt.utils.files.fopen(file_path, "a") as fp_:
            fp_.write("Added a line\n")
        # Commit the change
        self.assertNotIn(
            "ERROR",
            self.run_function(
                "git.commit",
                [self.repo, "Added a line to " + self.files[1]],
                filename=self.files[1],
            ),
        )
        # Switch to the second branch
        self.assertNotIn(
            "ERROR",
            self.run_function("git.checkout", [self.repo], rev=self.branches[1]),
        )
        # Perform the rebase. The commit should show a comment about
        # self.files[0] being modified, as that is the file that was modified
        # in the second branch in the setUp function
        ret = self.run_function("git.rebase", [self.repo], opts="-vvv")
        self.assertNotIn("ERROR", ret)
        self.assertNotIn("up to date", ret)

    # Test for git.remote_get is in test_remotes

    # Test for git.remote_set is in test_remotes

    @pytest.mark.slow_test
    def test_remotes(self):
        """
        Test setting a remote (git.remote_set), and getting a remote
        (git.remote_get and git.remotes)

        TODO: Properly mock fetching a remote (git.fetch), and build out more
        robust testing that confirms that the https auth bits work.
        """
        remotes = {
            "first": {"fetch": "/dev/null", "push": "/dev/null"},
            "second": {"fetch": "/dev/null", "push": "/dev/stdout"},
        }
        self.assertEqual(
            self.run_function(
                "git.remote_set", [self.repo, remotes["first"]["fetch"]], remote="first"
            ),
            remotes["first"],
        )
        self.assertEqual(
            self.run_function(
                "git.remote_set",
                [self.repo, remotes["second"]["fetch"]],
                remote="second",
                push_url=remotes["second"]["push"],
            ),
            remotes["second"],
        )
        self.assertEqual(self.run_function("git.remotes", [self.repo]), remotes)

    @pytest.mark.slow_test
    def test_reset(self):
        """
        Test git.reset

        TODO: Test more than just a hard reset
        """
        # Switch to the second branch
        self.assertTrue(
            "ERROR"
            not in self.run_function("git.checkout", [self.repo], rev=self.branches[1])
        )
        # Back up one commit. We should now be at the same revision as master
        self.run_function("git.reset", [self.repo], opts="--hard HEAD~1")
        # Get the SHA1 of current HEAD (remember, we're on the second branch)
        head_rev = self.run_function("git.revision", [self.repo], rev="HEAD")
        # Make sure revision is a 40-char string
        self.assertTrue(len(head_rev) == 40)
        # Get the master branch's SHA1
        master_rev = self.run_function("git.revision", [self.repo], rev="master")
        # Make sure revision is a 40-char string
        self.assertTrue(len(master_rev) == 40)
        # The two revisions should be the same
        self.assertEqual(head_rev, master_rev)

    @pytest.mark.slow_test
    def test_rev_parse(self):
        """
        Test git.rev_parse
        """
        # Using --abbrev-ref on HEAD will give us the current branch
        self.assertEqual(
            self.run_function(
                "git.rev_parse", [self.repo, "HEAD"], opts="--abbrev-ref"
            ),
            "master",
        )

    # Test for git.revision happens in test_merge_base

    @pytest.mark.slow_test
    def test_rm(self):
        """
        Test git.rm
        """
        single_file = self.files[0]
        entire_dir = self.dirs[1]
        # Remove a single file
        self.assertEqual(
            self.run_function("git.rm", [self.repo, single_file]),
            "rm '" + single_file + "'",
        )
        # Remove an entire dir
        expected = "\n".join(
            sorted("rm '" + os.path.join(entire_dir, x) + "'" for x in self.files)
        )
        if salt.utils.platform.is_windows():
            expected = expected.replace("\\", "/")
        self.assertEqual(
            self.run_function("git.rm", [self.repo, entire_dir], opts="-r"), expected
        )

    @pytest.mark.slow_test
    def test_stash(self):
        """
        Test git.stash

        # TODO: test more stash actions
        """
        file_path = os.path.join(self.repo, self.files[0])
        with salt.utils.files.fopen(file_path, "a") as fp_:
            fp_.write("Temp change to be stashed")
        self.assertTrue("ERROR" not in self.run_function("git.stash", [self.repo]))
        # List stashes
        ret = self.run_function("git.stash", [self.repo], action="list")
        self.assertTrue("ERROR" not in ret)
        self.assertTrue(len(ret.splitlines()) == 1)
        # Apply the stash
        self.assertTrue(
            "ERROR"
            not in self.run_function(
                "git.stash", [self.repo], action="apply", opts="stash@{0}"
            )
        )
        # Drop the stash
        self.assertTrue(
            "ERROR"
            not in self.run_function(
                "git.stash", [self.repo], action="drop", opts="stash@{0}"
            )
        )

    @pytest.mark.slow_test
    def test_status(self):
        """
        Test git.status
        """
        changes = {
            "modified": ["foo"],
            "new": ["thisisdefinitelyanewfile"],
            "deleted": ["bar"],
            "untracked": ["thisisalsoanewfile"],
        }
        for filename in changes["modified"]:
            with salt.utils.files.fopen(os.path.join(self.repo, filename), "a") as fp_:
                fp_.write("Added a line\n")
        for filename in changes["new"]:
            with salt.utils.files.fopen(os.path.join(self.repo, filename), "w") as fp_:
                fp_.write(
                    salt.utils.stringutils.to_str(
                        "This is a new file named {}.".format(filename)
                    )
                )
            # Stage the new file so it shows up as a 'new' file
            self.assertTrue(
                "ERROR" not in self.run_function("git.add", [self.repo, filename])
            )
        for filename in changes["deleted"]:
            self.run_function("git.rm", [self.repo, filename])
        for filename in changes["untracked"]:
            with salt.utils.files.fopen(os.path.join(self.repo, filename), "w") as fp_:
                fp_.write(
                    salt.utils.stringutils.to_str(
                        "This is a new file named {}.".format(filename)
                    )
                )
        self.assertEqual(self.run_function("git.status", [self.repo]), changes)

    # TODO: Add git.submodule test

    @pytest.mark.slow_test
    def test_symbolic_ref(self):
        """
        Test git.symbolic_ref
        """
        self.assertEqual(
            self.run_function("git.symbolic_ref", [self.repo, "HEAD"], opts="--quiet"),
            "refs/heads/master",
        )

    @pytest.mark.skipif(
        not _worktrees_supported(),
        reason="Git 2.5 or newer required for worktree support",
    )
    @pytest.mark.slow_test
    def test_worktree_add_rm(self):
        """
        This tests git.worktree_add, git.is_worktree, git.worktree_rm, and
        git.worktree_prune. Tests for 'git worktree list' are covered in
        tests.unit.modules.git_test.
        """
        # We don't need to enclose this comparison in a try/except, since the
        # decorator would skip this test if git is not installed and we'd never
        # get here in the first place.
        git_version = _git_version()
        if git_version >= LooseVersion("2.6.0"):
            worktree_add_prefix = "Preparing "
        else:
            worktree_add_prefix = "Enter "

        worktree_path = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        worktree_basename = os.path.basename(worktree_path)
        worktree_path2 = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        worktree_basename2 = os.path.basename(worktree_path2)

        # Even though this is Windows, git commands return a unix style path
        if salt.utils.platform.is_windows():
            worktree_path = worktree_path.replace("\\", "/")
            worktree_path2 = worktree_path2.replace("\\", "/")

        # Add the worktrees
        ret = self.run_function(
            "git.worktree_add",
            [self.repo, worktree_path],
        )
        self.assertTrue(worktree_add_prefix in ret)
        self.assertTrue(worktree_basename in ret)
        ret = self.run_function("git.worktree_add", [self.repo, worktree_path2])
        self.assertTrue(worktree_add_prefix in ret)
        self.assertTrue(worktree_basename2 in ret)
        # Check if this new path is a worktree
        self.assertTrue(self.run_function("git.is_worktree", [worktree_path]))
        # Check if the main repo is a worktree
        self.assertFalse(self.run_function("git.is_worktree", [self.repo]))
        # Check if a non-repo directory is a worktree
        empty_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.assertFalse(self.run_function("git.is_worktree", [empty_dir]))
        shutil.rmtree(empty_dir)
        # Remove the first worktree
        self.assertTrue(self.run_function("git.worktree_rm", [worktree_path]))
        # Prune the worktrees
        prune_message = (
            "Removing worktrees/{}: gitdir file points to non-existent location".format(
                worktree_basename
            )
        )
        # Test dry run output. It should match the same output we get when we
        # actually prune the worktrees.
        result = self.run_function("git.worktree_prune", [self.repo], dry_run=True)
        self.assertEqual(result, prune_message)
        # Test pruning for real, and make sure the output is the same
        self.assertEqual(
            self.run_function("git.worktree_prune", [self.repo]), prune_message
        )



"""
Testing Jinja filters availablilty via state system
"""
import logging
import os

import attr
import pytest

import salt.utils.files
import salt.utils.path

log = logging.getLogger(__name__)


@attr.s
class Filter:
    name = attr.ib()
    sls = attr.ib()
    expected = attr.ib(default=None)
    _exits = attr.ib(init=False, repr=False, factory=list)

    def check_skip(self, grains):
        pass

    def assert_result(self, changes):
        __tracebackhide__ = True
        assert changes
        if self.expected:
            if callable(self.expected):
                assert self.expected(changes)
            else:
                assert changes == self.expected

    def __call__(self, state_tree):
        self.state_tree = state_tree
        return self

    def __enter__(self):
        filter_sls = pytest.helpers.temp_file("filter.sls", self.sls, self.state_tree)
        filter_sls.__enter__()
        self._exits.append(filter_sls)
        return self

    def __exit__(self, *_):
        for exit_callback in self._exits:
            exit_callback.__exit__(*_)


@attr.s
class SkipOnWindowsFilter(Filter):
    def check_skip(self, grains):
        if grains["os"] == "Windows":
            pytest.skip("Skipped on windows")


@attr.s
class StartsWithFilter(Filter):
    def assert_result(self, changes):
        __tracebackhide__ = True
        assert changes
        assert changes["ret"]
        assert changes["ret"].startswith(self.expected)


@attr.s
class SortedFilter(Filter):
    def assert_result(self, changes):
        __tracebackhide__ = True
        assert changes
        assert changes["ret"]
        assert sorted(changes["ret"]) == sorted(self.expected["ret"])


@attr.s
class EmptyFileFilter(Filter):

    name = attr.ib(default="is_empty")
    sls = attr.ib(
        default="""
        {% set result = 'FPATH' | is_empty() %}
        test:
          module.run:
            - name: test.echo
            - text: {{ result }}
        """
    )

    def __enter__(self):
        empty_file = pytest.helpers.temp_file("empty-file", "", self.state_tree)
        fpath = empty_file.__enter__()
        self._exits.append(empty_file)
        self.sls = self.sls.replace("FPATH", str(fpath))
        return super().__enter__()


@attr.s
class TextFileFilter(Filter):

    name = attr.ib(default="is_text_file")
    sls = attr.ib(
        default="""
        {% set result = 'FPATH' | is_text_file() %}
        test:
          module.run:
            - name: test.echo
            - text: {{ result }}
        """
    )

    def __enter__(self):
        text_file = pytest.helpers.temp_file(
            "text-file", "This is a text file", self.state_tree
        )
        fpath = text_file.__enter__()
        self._exits.append(text_file)
        self.sls = self.sls.replace("FPATH", str(fpath))
        return super().__enter__()


@attr.s
class ListFilesFilter(SortedFilter):

    name = attr.ib(default="list_files")
    sls = attr.ib(
        default="""
        {% set result = 'FPATH' | list_files() %}
        test:
          module.run:
            - name: test.echo
            - text: {{ result }}
        """
    )

    def __enter__(self):
        text_file = pytest.helpers.temp_file(
            "foo/text-file", "This is a text file", self.state_tree
        )
        fpath = text_file.__enter__()
        self._exits.append(text_file)
        self.sls = self.sls.replace("FPATH", str(self.state_tree / "foo"))
        self.expected = {"ret": [str(self.state_tree / "foo"), str(fpath)]}
        return super().__enter__()


@attr.s
class FileHashsumFilter(Filter):

    name = attr.ib(default="file_hashsum")
    expected = attr.ib(
        default={
            "ret": "bfae4a86e38196ebccd4b9ef32454ff4271afa4ad539106de37d318591533873"
        }
    )
    sls = attr.ib(
        default="""
        {% set result = 'FPATH' | file_hashsum() %}
        test:
          module.run:
            - name: test.echo
            - text: {{ result }}
        """
    )

    def __enter__(self):
        text_file = pytest.helpers.temp_file(
            "text-file", "This is a text file", self.state_tree
        )
        fpath = text_file.__enter__()
        self._exits.append(text_file)
        self.sls = self.sls.replace("FPATH", str(fpath))
        return super().__enter__()


def _filter_id(value):
    return value.name


@pytest.fixture(
    params=[
        Filter(
            name="compare_dicts",
            expected={"ret": {"a": {"new": "c", "old": "b"}}},
            sls="""
            {% set dict_one = {'a': 'b', 'c': 'd'} %}
            {% set dict_two = {'a': 'c', 'c': 'd'} %}

            {% set result = dict_one | compare_dicts(dict_two) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result|tojson }}
            """,
        ),
        Filter(
            name="compare_lists",
            expected={"ret": {"old": ["b"]}},
            sls="""
            {% set list_one = ['a', 'b', 'c', 'd'] %}
            {% set list_two = ['a', 'c', 'd'] %}

            {% set result = list_one | compare_lists(list_two) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result|tojson }}
            """,
        ),
        Filter(
            name="json_decode_dict",
            expected={"ret": {"b'a'": "b'b'", "b'c'": "b'd'"}},
            sls="""
            {% set dict_one = {'a': 'b', 'c': 'd'} %}

            {% set result = dict_one | json_decode_dict() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="json_decode_list",
            expected={"ret": ["b'a'", "b'b'", "b'c'", "b'd'"]},
            sls="""
            {% set list_one = ['a', 'b', 'c', 'd'] %}

            {% set result = list_one | json_decode_list() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="json_encode_dict",
            expected={"ret": {"b'a'": "b'b'", "b'c'": "b'd'"}},
            sls="""
            {% set dict_one = {'a': 'b', 'c': 'd'} %}

            {% set result = dict_one | json_encode_dict() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="json_encode_list",
            expected={"ret": ["b'a'", "b'b'", "b'c'", "b'd'"]},
            sls="""
            {% set list_one = ['a', 'b', 'c', 'd'] %}

            {% set result = list_one | json_encode_list() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="exactly_n_true",
            expected={"ret": True},
            sls="""
            {% set list = [True, False, False, True] %}

            {% set result = list | exactly_n_true(2) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="exactly_one_true",
            expected={"ret": True},
            sls="""
            {% set list = [True, False, False, False] %}

            {% set result = list | exactly_one_true() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="is_iter",
            expected={"ret": True},
            sls="""
            {% set list = ['abcd','efgh','ijk','lmno','pqrs'] %}

            {% set result = list | is_iter() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="is_list",
            expected={"ret": True},
            sls="""
            {% set list = ['abcd','efgh','ijk','lmno','pqrs'] %}

            {% set result = list | is_list() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="mysql_to_dict",
            expected={
                "ret": {
                    "show processlist": {
                        "Info": "show processlist",
                        "db": "NULL",
                        "Host": "localhost",
                        "State": "init",
                        "Command": "Query",
                        "User": "root",
                        "Time": 0,
                        "Id": 7,
                    }
                }
            },
            sls="""
            {% set test_mysql_output =  ['+----+------+-----------+------+---------+------+-------+------------------+',
                                         '| Id | User | Host      | db   | Command | Time | State | Info             |',
                                         '+----+------+-----------+------+---------+------+-------+------------------+',
                                         '|  7 | root | localhost | NULL | Query   |    0 | init  | show processlist |',
                                         '+----+------+-----------+------+---------+------+-------+------------------+'] %}

            {% set result = test_mysql_output | mysql_to_dict('Info') %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result|tojson }}
            """,
        ),
        Filter(
            name="sorted_ignorecase",
            expected={"ret": ["Abcd", "efgh", "Ijk", "lmno", "Pqrs"]},
            sls="""
            {% set list = ['lmno','efgh','Ijk','Pqrs','Abcd'] %}

            {% set result = list | sorted_ignorecase() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="substring_in_list",
            expected={"ret": True},
            sls="""
            {% set string = 'lmno' %}
            {% set list = ['abcd','efgh','ijk','lmno','pqrs'] %}

            {% set result = string | substring_in_list(list) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="strftime",
            sls="""
            {% set result = none | strftime('%Y-%m-%d') %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="is_bin_file",
            expected={"ret": True},
            sls=r"""
            {% if grains['os'] == 'Windows' %}
              {% set result = 'c:\Windows\System32\cmd.exe' | is_bin_file() %}
            {% else %}
              {% set result = '/bin/ls' | is_bin_file() %}
            {% endif %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="base64_decode",
            expected={"ret": "Salt Rocks!"},
            sls="""
            {% set result = 'U2FsdCBSb2NrcyE=' | base64_decode() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="base64_encode",
            expected={"ret": "U2FsdCBSb2NrcyE="},
            sls="""
            {% set result = 'Salt Rocks!' | base64_encode() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        EmptyFileFilter(),
        TextFileFilter(),
        ListFilesFilter(),
        FileHashsumFilter(),
        Filter(
            name="hmac",
            expected={"ret": True},
            sls="""
            {% set result = 'Salt Rocks!' | hmac(shared_secret='topsecret', challenge_hmac='nMgLxwHPFyRgGfunkXXAI3Z/ZR4p5lmPTUjk2eGDqks=') %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="md5",
            expected={"ret": "85d6e71db655ee8e42c8b18475f0925f"},
            sls="""
            {% set result = 'Salt Rocks!' | md5() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="random_hash",
            sls="""
            {% set result = 9999999999 | random_hash() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="sha256",
            expected={
                "ret": "cce7fe00fd9cc6122fd3b2ed22feae215bcfe7ac4a7879d336c1993426a763fe"
            },
            sls="""
            {% set result = 'Salt Rocks!' | sha256() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="sha512",
            expected={
                "ret": "44d829491d8caa7039ad08a5b7fa9dd0f930138c614411c5326dd4755fdc9877ec6148219fccbe404139e7bb850e77237429d64f560c204f3697ab489fd8bfa5"
            },
            sls="""
            {% set result = 'Salt Rocks!' | sha512() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="http_query",
            expected={"ret": {}},
            sls="""
            {% set result = 'https://www.google.com' | http_query(test=True) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="avg",
            expected={"ret": 2.5},
            sls="""
            {% set result = [1, 2, 3, 4] | avg() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="difference",
            expected={"ret": [1, 3]},
            sls="""
            {% set result = [1, 2, 3, 4] | difference([2, 4, 6]) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="intersect",
            expected={"ret": [2, 4]},
            sls="""
            {% set result = [1, 2, 3, 4] | intersect([2, 4, 6]) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="max",
            expected={"ret": 4},
            sls="""
            {% set result = [1, 2, 3, 4] | max() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="min",
            expected={"ret": 1},
            sls="""
            {% set result = [1, 2, 3, 4] | min() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="quote",
            expected={"ret": "Salt Rocks!"},
            sls="""
            {% set result = 'Salt Rocks!' | quote() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="regex_escape",
            expected={"ret": r"Salt\ Rocks"},
            sls="""
            {% set result = 'Salt Rocks' | regex_escape() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="regex_match",
            expected={"ret": "('a', 'd')"},
            sls="""
            {% set result = 'abcd' | regex_match('^(.*)BC(.*)$', ignorecase=True) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="regex_replace",
            expected={"ret": "lets__replace__spaces"},
            sls=r"""
            {% set result = 'lets replace spaces' | regex_replace('\s+', '__') %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="regex_search",
            expected={"ret": "('a', 'd')"},
            sls="""
            {% set result = 'abcd' | regex_search('^(.*)BC(.*)$', ignorecase=True) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="sequence",
            expected={"ret": ["Salt Rocks!"]},
            sls="""
            {% set result = 'Salt Rocks!' | sequence() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="skip",
            expected={"ret": None},
            sls="""
            {% set result = 'Salt Rocks!' | skip() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="symmetric_difference",
            expected={"ret": [1, 3, 6]},
            sls="""
            {% set result = [1, 2, 3, 4] | symmetric_difference([2, 4, 6]) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="to_bool",
            expected={"ret": True},
            sls="""
            {% set result = 'yes' | to_bool() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="union",
            expected={"ret": [1, 2, 3, 4, 6]},
            sls="""
            {% set result = [1, 2, 3, 4] | union([2, 4, 6]) %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="unique",
            expected={"ret": ["a", "b", "c"]},
            sls="""
            {% set result = ['a', 'b', 'c', 'a', 'b'] | unique() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="uuid",
            expected={"ret": "799192d9-7f32-5227-a45f-dfeb4a34e06f"},
            sls="""
            {% set result = 'Salt Rocks!' | uuid() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        StartsWithFilter(
            name="gen_mac",
            expected="AC:DE:48:",
            sls="""
            {% set result = 'AC:DE:48' | gen_mac() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        SortedFilter(
            name="ipaddr",
            expected={"ret": ["127.0.0.1", "::1"]},
            sls="""
            {% set result = 'Salt Rocks!' | uuid() %}
            {% set result = ['127.0.0.1', '::1', 'random_junk'] | ipaddr() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="ip_host",
            expected={"ret": "192.168.0.12/24"},
            sls="""
            {% set result = '192.168.0.12/24' | ip_host() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="ipv4",
            expected={"ret": ["127.0.0.1"]},
            sls="""
            {% set result = ['127.0.0.1', '::1'] | ipv4() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="ipv6",
            expected={"ret": ["::1"]},
            sls="""
            {% set result = ['127.0.0.1', '::1'] | ipv6() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="is_ip",
            expected={"ret": True},
            sls="""
            {% set result = '127.0.0.1' | is_ip() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="is_ipv4",
            expected={"ret": True},
            sls="""
            {% set result = '127.0.0.1' | is_ipv4() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="is_ipv6",
            expected={"ret": True},
            sls="""
            {% set result = '::1' | is_ipv6() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="network_hosts",
            expected={"ret": ["192.168.1.1", "192.168.1.2"]},
            sls="""
            {% set result = '192.168.1.0/30' | network_hosts() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="network_size",
            expected={"ret": 16},
            sls="""
            {% set result = '192.168.1.0/28' | network_size() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="path_join",
            expected={"ret": os.path.sep + os.path.join("a", "b", "c", "d")},
            sls="""
            {% set result = '/a/b/c' | path_join('d') %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="which",
            expected={"ret": salt.utils.path.which("which")},
            sls="""
            {% set result = 'which' | which() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="contains_whitespace",
            expected={"ret": True},
            sls="""
            {% set result = 'This string has whitespace' | contains_whitespace() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="is_hex",
            expected={"ret": True},
            sls="""
            {% set result = '0x2a' | is_hex() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="random_str",
            expected=lambda x: len(x["ret"]) == 32,
            sls="""
            {% set result = 32 | random_str() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="to_num",
            expected={"ret": 42},
            sls="""
            {% set result = '42' | to_num() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="check_whitelist_blacklist",
            expected={"ret": True},
            sls="""
            {% set result = 'string' | check_whitelist_blacklist(whitelist='string') %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        SkipOnWindowsFilter(
            name="get_uid",
            expected={"ret": 0},
            sls="""
            {% set result = 'root' | get_uid() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="yaml_dquote",
            expected={"ret": "A double-quoted string in YAML"},
            sls="""
            {% set result = "A double-quoted string in YAML" | yaml_dquote() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="yaml_squote",
            expected={"ret": "A single-quoted string in YAML"},
            sls="""
            {% set result = 'A single-quoted string in YAML' | yaml_squote() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="yaml_encode",
            expected={"ret": "An encoded string in YAML"},
            sls="""
            {% set result = "An encoded string in YAML" | yaml_encode() %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="yaml",
            expected={"ret": {"Question": "Quieres Café?"}},
            sls="""
            {% set result = {"Question": "Quieres Café?"} | yaml %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
        Filter(
            name="json",
            expected={"ret": {"Question": "Quieres Café?"}},
            sls="""
            {% set result = {"Question": "Quieres Café?"} | json %}
            test:
              module.run:
                - name: test.echo
                - text: {{ result }}
            """,
        ),
    ],
    ids=_filter_id,
)
def filter(request):
    return request.param


def test_filter(state, state_tree, filter, grains):
    filter.check_skip(grains)
    with filter(state_tree):
        ret = state.sls("filter")
        log.debug("state.sls returned: %s", ret)
        assert not ret.failed
        for state_result in ret:
            assert state_result.result is True
            filter.assert_result(state_result.changes)



"""
NET Finder
==========

.. versionadded:: 2017.7.0

A runner to find network details easily and fast.
It's smart enough to know what you are looking for.

Configuration
-------------

- Minion (proxy) config

    To have the complete features, one needs to add the following mine configuration in the minion (proxy) config file:

    .. code-block:: yaml

        mine_functions:
          net.ipaddrs: []
          net.lldp: []
          net.mac: []
          net.arp: []
          net.interfaces: []

    Which instructs Salt to cache the data returned by the NAPALM-functions.
    While they are not mandatory, the less functions configured, the less details will be found by the runner.

    How often the mines are refreshed, can be specified using:

    .. code-block:: yaml

        mine_interval: <X minutes>

- Master config

    By default the following options can be configured on the master.
    They are not necessary, but available in case the user has different requirements.

    target: ``*``
        From what minions will collect the mine data. Default: ``*`` (collect from all minions).

    expr_form: ``glob``
        Minion matching expression form. Default: ``glob``.

    ignore_interfaces
        A list of interfaces name to ignore. By default will consider all interfaces.

    display: ``True``
        Display on the screen or return structured object? Default: ``True`` (return on the CLI).

    outputter: ``table``
        Specify the outputter name when displaying on the CLI. Default: :mod:`table <salt.output.napalm_bgp>`.

    Configuration example:

    .. code-block:: yaml

        runners:
          net.find:
            target: 'edge*'
            expr_form: 'glob'
            ignore_interfaces:
              - lo0
              - em1
              - jsrv
              - fxp0
            outputter: yaml
"""

import salt.output
import salt.utils.network

try:
    # pylint: disable=no-name-in-module
    from napalm.base import helpers as napalm_helpers
    from netaddr import IPNetwork  # netaddr is already required by napalm
    from netaddr.core import AddrFormatError

    # pylint: enable=no-name-in-module

    HAS_NAPALM = True
except ImportError:
    HAS_NAPALM = False

# -----------------------------------------------------------------------------
# module properties
# -----------------------------------------------------------------------------

_DEFAULT_TARGET = "*"
_DEFAULT_EXPR_FORM = "glob"
_DEFAULT_IGNORE_INTF = []
# 'lo0', 'em1', 'em0', 'jsrv', 'fxp0'
_DEFAULT_DISPLAY = True
_DEFAULT_OUTPUTTER = "table"


# -----------------------------------------------------------------------------
# global variables
# -----------------------------------------------------------------------------

# will cache several details to avoid loading them several times from the mines.
_CACHE = {}

# -----------------------------------------------------------------------------
# helper functions -- will not be exported
# -----------------------------------------------------------------------------

# Define the module's virtual name
__virtualname__ = "net"


def __virtual__():
    if HAS_NAPALM:
        return __virtualname__
    return (False, "The napalm module could not be imported")


def _get_net_runner_opts():
    """
    Return the net.find runner options.
    """
    runner_opts = __opts__.get("runners", {}).get("net.find", {})
    return {
        "target": runner_opts.get("target", _DEFAULT_TARGET),
        "expr_form": runner_opts.get("expr_form", _DEFAULT_EXPR_FORM),
        "ignore_interfaces": runner_opts.get("ignore_interfaces", _DEFAULT_IGNORE_INTF),
        "display": runner_opts.get("display", _DEFAULT_DISPLAY),
        "outputter": runner_opts.get("outputter", _DEFAULT_OUTPUTTER),
    }


def _get_mine(fun):
    """
    Return the mine function from all the targeted minions.
    Just a small helper to avoid redundant pieces of code.
    """
    if fun in _CACHE and _CACHE[fun]:
        return _CACHE[fun]
    net_runner_opts = _get_net_runner_opts()
    _CACHE[fun] = __salt__["mine.get"](
        net_runner_opts.get("target"), fun, tgt_type=net_runner_opts.get("expr_form")
    )
    return _CACHE[fun]


def _display_runner(rows, labels, title, display=_DEFAULT_DISPLAY):
    """
    Display or return the rows.
    """
    if display:
        net_runner_opts = _get_net_runner_opts()
        if net_runner_opts.get("outputter") == "table":
            ret = salt.output.out_format(
                {"rows": rows, "labels": labels},
                "table",
                __opts__,
                title=title,
                rows_key="rows",
                labels_key="labels",
            )
        else:
            ret = salt.output.out_format(
                rows, net_runner_opts.get("outputter"), __opts__
            )
        print(ret)
    else:
        return rows


def _get_network_obj(addr):
    """
    Try to convert a string into a valid IP Network object.
    """
    ip_netw = None
    try:
        ip_netw = IPNetwork(addr)
    except AddrFormatError:
        return ip_netw
    return ip_netw


def _find_interfaces_ip(mac):
    """
    Helper to search the interfaces IPs using the MAC address.
    """
    try:
        mac = napalm_helpers.convert(napalm_helpers.mac, mac)
    except AddrFormatError:
        return ("", "", [])

    all_interfaces = _get_mine("net.interfaces")
    all_ipaddrs = _get_mine("net.ipaddrs")

    for device, device_interfaces in all_interfaces.items():
        if not device_interfaces.get("result", False):
            continue
        for interface, interface_details in device_interfaces.get("out", {}).items():
            try:
                interface_mac = napalm_helpers.convert(
                    napalm_helpers.mac, interface_details.get("mac_address")
                )
            except AddrFormatError:
                continue
            if mac != interface_mac:
                continue
            interface_ipaddrs = (
                all_ipaddrs.get(device, {}).get("out", {}).get(interface, {})
            )
            ip_addresses = interface_ipaddrs.get("ipv4", {})
            ip_addresses.update(interface_ipaddrs.get("ipv6", {}))
            interface_ips = [
                "{}/{}".format(ip_addr, addr_details.get("prefix_length", "32"))
                for ip_addr, addr_details in ip_addresses.items()
            ]
            return device, interface, interface_ips

    return ("", "", [])


def _find_interfaces_mac(ip):  # pylint: disable=invalid-name
    """
    Helper to get the interfaces hardware address using the IP Address.
    """
    all_interfaces = _get_mine("net.interfaces")
    all_ipaddrs = _get_mine("net.ipaddrs")

    for device, device_ipaddrs in all_ipaddrs.items():
        if not device_ipaddrs.get("result", False):
            continue
        for interface, interface_ipaddrs in device_ipaddrs.get("out", {}).items():
            ip_addresses = set(interface_ipaddrs.get("ipv4", {}).keys())
            ip_addresses.update(set(interface_ipaddrs.get("ipv6", {}).keys()))
            for ipaddr in ip_addresses:
                if ip != ipaddr:
                    continue
                interface_mac = (
                    all_interfaces.get(device, {})
                    .get("out", {})
                    .get(interface, {})
                    .get("mac_address", "")
                )
                return device, interface, interface_mac

    return ("", "", "")


# -----------------------------------------------------------------------------
# callable functions
# -----------------------------------------------------------------------------


def interfaces(
    device=None,
    interface=None,
    title=None,
    pattern=None,
    ipnet=None,
    best=True,
    display=_DEFAULT_DISPLAY,
):
    """
    Search for interfaces details in the following mine functions:

    - net.interfaces
    - net.ipaddrs

    Optional arguments:

    device
        Return interface data from a certain device only.

    interface
        Return data selecting by interface name.

    pattern
        Return interfaces that contain a certain pattern in their description.

    ipnet
        Return interfaces whose IP networks associated include this IP network.

    best: ``True``
        When ``ipnet`` is specified, this argument says if the runner should return only the best match
        (the output will contain at most one row). Default: ``True`` (return only the best match).

    display: True
        Display on the screen or return structured object? Default: ``True`` (return on the CLI).

    title
        Display a custom title for the table.

    CLI Example:

    .. code-block:: bash

        $ sudo salt-run net.interfaces interface=vt-0/0/10

    Output Example:

    .. code-block:: text

        Details for interface xe-0/0/0
        _________________________________________________________________________________________________________________
        |    Device    | Interface | Interface Description |  UP  | Enabled | Speed [Mbps] | MAC Address | IP Addresses |
        _________________________________________________________________________________________________________________
        | edge01.bjm01 | vt-0/0/10 |                       | True |   True  |     1000     |             |              |
        _________________________________________________________________________________________________________________
        | edge01.flw01 | vt-0/0/10 |                       | True |   True  |     1000     |             |              |
        _________________________________________________________________________________________________________________
        | edge01.pos01 | vt-0/0/10 |                       | True |   True  |     1000     |             |              |
        _________________________________________________________________________________________________________________
        | edge01.oua01 | vt-0/0/10 |                       | True |   True  |     1000     |             |              |
        _________________________________________________________________________________________________________________
    """

    def _ipnet_belongs(net):
        """
        Helper to tell if a IP address or network belong to a certain network.
        """
        if net == "0.0.0.0/0":
            return False
        net_obj = _get_network_obj(net)
        if not net_obj:
            return False
        return ipnet in net_obj or net_obj in ipnet

    labels = {
        "device": "Device",
        "interface": "Interface",
        "interface_description": "Interface Description",
        "is_up": "UP",
        "is_enabled": "Enabled",
        "speed": "Speed [Mbps]",
        "mac": "MAC Address",
        "ips": "IP Addresses",
    }
    rows = []

    net_runner_opts = _get_net_runner_opts()

    if pattern:
        title = (
            'Pattern "{}" found in the description of the following interfaces'.format(
                pattern
            )
        )
    if not title:
        title = "Details"
        if interface:
            title += " for interface {}".format(interface)
        else:
            title += " for all interfaces"
        if device:
            title += " on device {}".format(device)
        if ipnet:
            title += " that include network {net}".format(net=str(ipnet))
            if best:
                title += " - only best match returned"

    all_interfaces = _get_mine("net.interfaces")
    all_ipaddrs = _get_mine("net.ipaddrs")

    if device:
        all_interfaces = {device: all_interfaces.get(device, {})}

    if ipnet and not isinstance(ipnet, IPNetwork):
        ipnet = _get_network_obj(ipnet)

    best_row = {}
    best_net_match = IPNetwork("0.0.0.0/0")
    for (
        device,
        net_interfaces_out,
    ) in all_interfaces.items():  # pylint: disable=too-many-nested-blocks
        if not net_interfaces_out:
            continue
        if not net_interfaces_out.get("result", False):
            continue
        selected_device_interfaces = net_interfaces_out.get("out", {})
        if interface:
            selected_device_interfaces = {
                interface: selected_device_interfaces.get(interface, {})
            }
        for interface_name, interface_details in selected_device_interfaces.items():
            if not interface_details:
                continue
            if ipnet and interface_name in net_runner_opts.get("ignore_interfaces"):
                continue
            interface_description = interface_details.get("description", "") or ""
            if pattern:
                if pattern.lower() not in interface_description.lower():
                    continue
            if not all_ipaddrs.get(device, {}).get("result", False):
                continue
            ips = []
            device_entry = {
                "device": device,
                "interface": interface_name,
                "interface_description": interface_description,
                "is_up": (interface_details.get("is_up", "") or ""),
                "is_enabled": (interface_details.get("is_enabled", "") or ""),
                "speed": (interface_details.get("speed", "") or ""),
                "mac": napalm_helpers.convert(
                    napalm_helpers.mac, (interface_details.get("mac_address", "") or "")
                ),
                "ips": [],
            }
            intf_entry_found = False
            for intrf, interface_ips in (
                all_ipaddrs.get(device, {}).get("out", {}).items()
            ):
                if intrf.split(".")[0] == interface_name:
                    ip_addresses = interface_ips.get("ipv4", {})  # all IPv4 addresses
                    ip_addresses.update(
                        interface_ips.get("ipv6", {})
                    )  # and all IPv6 addresses
                    ips = [
                        "{}/{}".format(ip_addr, addr_details.get("prefix_length", "32"))
                        for ip_addr, addr_details in ip_addresses.items()
                    ]
                    interf_entry = {}
                    interf_entry.update(device_entry)
                    interf_entry["ips"] = ips
                    if display:
                        interf_entry["ips"] = "\n".join(interf_entry["ips"])
                    if ipnet:
                        inet_ips = [
                            str(ip) for ip in ips if _ipnet_belongs(ip)
                        ]  # filter and get only IP include ipnet
                        if inet_ips:  # if any
                            if best:
                                # determine the global best match
                                compare = [best_net_match]
                                compare.extend(list(map(_get_network_obj, inet_ips)))
                                new_best_net_match = max(compare)
                                if new_best_net_match != best_net_match:
                                    best_net_match = new_best_net_match
                                    best_row = interf_entry
                            else:
                                # or include all
                                intf_entry_found = True
                                rows.append(interf_entry)
                    else:
                        intf_entry_found = True
                        rows.append(interf_entry)
            if not intf_entry_found and not ipnet:
                interf_entry = {}
                interf_entry.update(device_entry)
                if display:
                    interf_entry["ips"] = ""
                rows.append(interf_entry)

    if ipnet and best and best_row:
        rows = [best_row]

    return _display_runner(rows, labels, title, display=display)


def findarp(
    device=None, interface=None, mac=None, ip=None, display=_DEFAULT_DISPLAY
):  # pylint: disable=invalid-name
    """
    Search for entries in the ARP tables using the following mine functions:

    - net.arp

    Optional arguments:

    device
        Return interface data from a certain device only.

    interface
        Return data selecting by interface name.

    mac
        Search using a specific MAC Address.

    ip
        Search using a specific IP Address.

    display: ``True``
        Display on the screen or return structured object? Default: ``True``, will return on the CLI.

    CLI Example:

    .. code-block:: bash

        $ sudo salt-run net.findarp mac=8C:60:0F:78:EC:41

    Output Example:

    .. code-block:: text

        ARP Entries for MAC 8C:60:0F:78:EC:41
        ________________________________________________________________________________
        |    Device    |     Interface     |        MAC        |       IP      |  Age  |
        ________________________________________________________________________________
        | edge01.bjm01 | irb.171 [ae0.171] | 8C:60:0F:78:EC:41 | 172.172.17.19 | 956.0 |
        ________________________________________________________________________________
    """
    labels = {
        "device": "Device",
        "interface": "Interface",
        "mac": "MAC",
        "ip": "IP",
        "age": "Age",
    }
    rows = []

    all_arp = _get_mine("net.arp")

    title = "ARP Entries"
    if device:
        title += " on device {device}".format(device=device)
    if interface:
        title += " on interface {interf}".format(interf=interface)
    if ip:
        title += " for IP {ip}".format(ip=ip)
    if mac:
        title += " for MAC {mac}".format(mac=mac)

    if device:
        all_arp = {device: all_arp.get(device)}

    for device, device_arp in all_arp.items():
        if not device_arp:
            continue
        if not device_arp.get("result", False):
            continue
        arp_table = device_arp.get("out", [])
        for arp_entry in arp_table:
            if (
                (mac and arp_entry.get("mac", "").lower() == mac.lower())
                or (  # pylint: disable=too-many-boolean-expressions
                    interface and interface in arp_entry.get("interface", "")
                )
                or (
                    ip
                    and napalm_helpers.convert(
                        napalm_helpers.ip, arp_entry.get("ip", "")
                    )
                    == napalm_helpers.convert(napalm_helpers.ip, ip)
                )
            ):
                rows.append(
                    {
                        "device": device,
                        "interface": arp_entry.get("interface"),
                        "mac": napalm_helpers.convert(
                            napalm_helpers.mac, arp_entry.get("mac")
                        ),
                        "ip": napalm_helpers.convert(
                            napalm_helpers.ip, arp_entry.get("ip")
                        ),
                        "age": arp_entry.get("age"),
                    }
                )

    return _display_runner(rows, labels, title, display=display)


def findmac(device=None, mac=None, interface=None, vlan=None, display=_DEFAULT_DISPLAY):
    """
    Search in the MAC Address tables, using the following mine functions:

    - net.mac

    Optional arguments:

    device
        Return interface data from a certain device only.

    interface
        Return data selecting by interface name.

    mac
        Search using a specific MAC Address.

    vlan
        Search using a VLAN ID.

    display: ``True``
        Display on the screen or return structured object? Default: ``True``, will return on the CLI.

    CLI Example:

    .. code-block:: bash

        $ sudo salt-run net.findmac mac=8C:60:0F:78:EC:41

    Output Example:

    .. code-block:: text

        MAC Address(es)
        _____________________________________________________________________________________________
        |    Device    | Interface |        MAC        | VLAN | Static | Active | Moves | Last move |
        _____________________________________________________________________________________________
        | edge01.bjm01 |  ae0.171  | 8C:60:0F:78:EC:41 | 171  | False  |  True  |   0   |    0.0    |
        _____________________________________________________________________________________________
    """
    labels = {
        "device": "Device",
        "interface": "Interface",
        "mac": "MAC",
        "vlan": "VLAN",
        "static": "Static",
        "active": "Active",
        "moves": "Moves",
        "last_move": "Last Move",
    }
    rows = []

    all_mac = _get_mine("net.mac")

    title = "MAC Address(es)"
    if device:
        title += " on device {device}".format(device=device)
    if interface:
        title += " on interface {interf}".format(interf=interface)
    if vlan:
        title += " on VLAN {vlan}".format(vlan=vlan)

    if device:
        all_mac = {device: all_mac.get(device)}

    for device, device_mac in all_mac.items():
        if not device_mac:
            continue
        if not device_mac.get("result", False):
            continue
        mac_table = device_mac.get("out", [])
        for mac_entry in mac_table:
            if (
                (
                    mac
                    and napalm_helpers.convert(  # pylint: disable=too-many-boolean-expressions
                        napalm_helpers.mac, mac_entry.get("mac", "")
                    )
                    == napalm_helpers.convert(napalm_helpers.mac, mac)
                )
                or (interface and interface in mac_entry.get("interface", ""))
                or (vlan and str(mac_entry.get("vlan", "")) == str(vlan))
            ):
                rows.append(
                    {
                        "device": device,
                        "interface": mac_entry.get("interface"),
                        "mac": napalm_helpers.convert(
                            napalm_helpers.mac, mac_entry.get("mac")
                        ),
                        "vlan": mac_entry.get("vlan"),
                        "static": mac_entry.get("static"),
                        "active": mac_entry.get("active"),
                        "moves": mac_entry.get("moves"),
                        "last_move": mac_entry.get("last_move"),
                    }
                )

    return _display_runner(rows, labels, title, display=display)


def lldp(
    device=None,
    interface=None,
    title=None,
    pattern=None,
    chassis=None,
    display=_DEFAULT_DISPLAY,
):
    """
    Search in the LLDP neighbors, using the following mine functions:

    - net.lldp

    Optional arguments:

    device
        Return interface data from a certain device only.

    interface
        Return data selecting by interface name.

    pattern
        Return LLDP neighbors that have contain this pattern in one of the following fields:

        - Remote Port ID
        - Remote Port Description
        - Remote System Name
        - Remote System Description

    chassis
        Search using a specific Chassis ID.

    display: ``True``
        Display on the screen or return structured object? Default: ``True`` (return on the CLI).

    display: ``True``
        Display on the screen or return structured object? Default: ``True`` (return on the CLI).

    title
        Display a custom title for the table.

    CLI Example:

    .. code-block:: bash

        $ sudo salt-run net.lldp pattern=Ethernet1/48

    Output Example:

    .. code-block:: text

        Pattern "Ethernet1/48" found in one of the following LLDP details
        _________________________________________________________________________________________________________________________________________________________________________________________
        |    Device    | Interface | Parent Interface | Remote Chassis ID | Remote Port ID | Remote Port Description |   Remote System Name   |            Remote System Description            |
        _________________________________________________________________________________________________________________________________________________________________________________________
        | edge01.bjm01 |  xe-2/3/4 |       ae0        | 8C:60:4F:3B:52:19 |                |       Ethernet1/48      | edge05.bjm01.dummy.net |   Cisco NX-OS(tm) n6000, Software (n6000-uk9),  |
        |              |           |                  |                   |                |                         |                        | Version 7.3(0)N7(5), RELEASE SOFTWARE Copyright |
        |              |           |                  |                   |                |                         |                        |  (c) 2002-2012 by Cisco Systems, Inc. Compiled  |
        |              |           |                  |                   |                |                         |                        |                2/17/2016 22:00:00               |
        _________________________________________________________________________________________________________________________________________________________________________________________
        | edge01.flw01 |  xe-1/2/3 |       ae0        | 8C:60:4F:1A:B4:22 |                |       Ethernet1/48      | edge05.flw01.dummy.net |   Cisco NX-OS(tm) n6000, Software (n6000-uk9),  |
        |              |           |                  |                   |                |                         |                        | Version 7.3(0)N7(5), RELEASE SOFTWARE Copyright |
        |              |           |                  |                   |                |                         |                        |  (c) 2002-2012 by Cisco Systems, Inc. Compiled  |
        |              |           |                  |                   |                |                         |                        |                2/17/2016 22:00:00               |
        _________________________________________________________________________________________________________________________________________________________________________________________
        | edge01.oua01 |  xe-0/1/2 |       ae1        | 8C:60:4F:51:A4:22 |                |       Ethernet1/48      | edge05.oua01.dummy.net |   Cisco NX-OS(tm) n6000, Software (n6000-uk9),  |
        |              |           |                  |                   |                |                         |                        | Version 7.3(0)N7(5), RELEASE SOFTWARE Copyright |
        |              |           |                  |                   |                |                         |                        |  (c) 2002-2012 by Cisco Systems, Inc. Compiled  |
        |              |           |                  |                   |                |                         |                        |                2/17/2016 22:00:00               |
        _________________________________________________________________________________________________________________________________________________________________________________________
    """
    all_lldp = _get_mine("net.lldp")

    labels = {
        "device": "Device",
        "interface": "Interface",
        "parent_interface": "Parent Interface",
        "remote_chassis_id": "Remote Chassis ID",
        "remote_port_id": "Remote Port ID",
        "remote_port_desc": "Remote Port Description",
        "remote_system_name": "Remote System Name",
        "remote_system_desc": "Remote System Description",
    }
    rows = []

    if pattern:
        title = 'Pattern "{}" found in one of the following LLDP details'.format(
            pattern
        )
    if not title:
        title = "LLDP Neighbors"
        if interface:
            title += " for interface {}".format(interface)
        else:
            title += " for all interfaces"
        if device:
            title += " on device {}".format(device)
        if chassis:
            title += " having Chassis ID {}".format(chassis)

    if device:
        all_lldp = {device: all_lldp.get(device)}

    for device, device_lldp in all_lldp.items():
        if not device_lldp:
            continue
        if not device_lldp.get("result", False):
            continue
        lldp_interfaces = device_lldp.get("out", {})
        if interface:
            lldp_interfaces = {interface: lldp_interfaces.get(interface, [])}
        for intrf, interface_lldp in lldp_interfaces.items():
            if not interface_lldp:
                continue
            for lldp_row in interface_lldp:
                rsn = lldp_row.get("remote_system_name", "") or ""
                rpi = lldp_row.get("remote_port_id", "") or ""
                rsd = lldp_row.get("remote_system_description", "") or ""
                rpd = lldp_row.get("remote_port_description", "") or ""
                rci = lldp_row.get("remote_chassis_id", "") or ""
                if pattern:
                    ptl = pattern.lower()
                    if not (
                        (ptl in rsn.lower())
                        or (ptl in rsd.lower())
                        or (ptl in rpd.lower())
                        or (ptl in rci.lower())
                    ):
                        # nothing matched, let's move on
                        continue
                if chassis:
                    if napalm_helpers.convert(
                        napalm_helpers.mac, rci
                    ) != napalm_helpers.convert(napalm_helpers.mac, chassis):
                        continue
                rows.append(
                    {
                        "device": device,
                        "interface": intrf,
                        "parent_interface": (
                            lldp_row.get("parent_interface", "") or ""
                        ),
                        "remote_chassis_id": napalm_helpers.convert(
                            napalm_helpers.mac, rci
                        ),
                        "remote_port_id": rpi,
                        "remote_port_descr": rpd,
                        "remote_system_name": rsn,
                        "remote_system_descr": rsd,
                    }
                )

    return _display_runner(rows, labels, title, display=display)


def find(addr, best=True, display=_DEFAULT_DISPLAY):
    """
    Search in all possible entities (Interfaces, MAC tables, ARP tables, LLDP neighbors),
    using the following mine functions:

    - net.mac
    - net.arp
    - net.lldp
    - net.ipaddrs
    - net.interfaces

    This function has the advantage that it knows where to look, but the output might
    become quite long as returns all possible matches.

    Optional arguments:

    best: ``True``
        Return only the best match with the interfaces IP networks
        when the saerching pattern is a valid IP Address or Network.

    display: ``True``
        Display on the screen or return structured object? Default: ``True`` (return on the CLI).

    CLI Example:

    .. code-block:: bash

        $ sudo salt-run net.find 10.10.10.7

    Output Example:

    .. code-block:: text

        Details for all interfaces that include network 10.10.10.7/32 - only best match returned
        ________________________________________________________________________________________________________________________
        |    Device    | Interface | Interface Description |  UP  | Enabled | Speed [Mbps] |    MAC Address    |  IP Addresses |
        ________________________________________________________________________________________________________________________
        | edge01.flw01 |    irb    |                       | True |   True  |      -1      | 5C:5E:AB:AC:52:B4 | 10.10.10.1/22 |
        ________________________________________________________________________________________________________________________

        ARP Entries for IP 10.10.10.7
        _____________________________________________________________________________
        |    Device    |     Interface     |        MAC        |     IP     |  Age  |
        _____________________________________________________________________________
        | edge01.flw01 | irb.349 [ae0.349] | 2C:60:0C:2A:4C:0A | 10.10.10.7 | 832.0 |
        _____________________________________________________________________________
    """
    if not addr:
        if display:
            print("Please type a valid MAC/IP Address / Device / Interface / VLAN")
        return {}

    device = ""
    interface = ""
    mac = ""
    ip = ""  # pylint: disable=invalid-name
    ipnet = None

    results = {
        "int_net": [],
        "int_descr": [],
        "int_name": [],
        "int_ip": [],
        "int_mac": [],
        "int_device": [],
        "lldp_descr": [],
        "lldp_int": [],
        "lldp_device": [],
        "lldp_mac": [],
        "lldp_device_int": [],
        "mac_device": [],
        "mac_int": [],
        "arp_device": [],
        "arp_int": [],
        "arp_mac": [],
        "arp_ip": [],
    }

    if isinstance(addr, int):
        results["mac"] = findmac(vlan=addr, display=display)
        if not display:
            return results
        else:
            return None

    try:
        mac = napalm_helpers.convert(napalm_helpers.mac, addr)
    except IndexError:
        # no problem, let's keep searching
        pass
    if salt.utils.network.is_ipv6(addr):
        mac = False
    if not mac:
        try:
            ip = napalm_helpers.convert(
                napalm_helpers.ip, addr
            )  # pylint: disable=invalid-name
        except ValueError:
            pass
        ipnet = _get_network_obj(addr)
        if ipnet:
            results["int_net"] = interfaces(ipnet=ipnet, best=best, display=display)
        if not (ipnet or ip):
            # search in all possible places
            # display all interfaces details
            results["int_descr"] = interfaces(pattern=addr, display=display)
            results["int_name"] = interfaces(interface=addr, display=display)
            results["int_device"] = interfaces(device=addr, display=display)
            # search in LLDP details
            results["lldp_descr"] = lldp(pattern=addr, display=display)
            results["lldp_int"] = lldp(interface=addr, display=display)
            results["lldp_device"] = lldp(device=addr, display=display)
            # search in MAC Address tables
            results["mac_device"] = findmac(device=addr, display=display)
            results["mac_int"] = findmac(interface=addr, display=display)
            # search in ARP tables
            results["arp_device"] = findarp(device=addr, display=display)
            results["arp_int"] = findarp(interface=addr, display=display)
            if not display:
                return results
    if mac:
        results["int_descr"] = findmac(mac=mac, display=display)
        results["arp_mac"] = findarp(mac=mac, display=display)
        results["lldp_mac"] = lldp(chassis=mac, display=display)
    if ip:
        results["arp_ip"] = findarp(ip=ip, display=display)

    # let's search in Interfaces

    if mac:
        device, interface, ips = _find_interfaces_ip(mac)
        ip = ", ".join(ips)  # pylint: disable=invalid-name
        if device and interface:
            title = "Interface {interface} on {device} has the physical address ({mac})".format(
                interface=interface, device=device, mac=mac
            )
            results["int_mac"] = interfaces(
                device=device, interface=interface, title=title, display=display
            )

    elif ip:
        device, interface, mac = _find_interfaces_mac(ip)
        if device and interface:
            title = (
                "IP Address {ip} is set for interface {interface}, on {device}".format(
                    interface=interface, device=device, ip=ip
                )
            )
            results["int_ip"] = interfaces(
                device=device, interface=interface, title=title, display=display
            )

    if device and interface:
        results["lldp_device_int"] = lldp(device, interface, display=display)

    if not display:
        return results


def multi_find(*patterns, **kwargs):
    """
    Execute multiple search tasks.
    This function is based on the `find` function.
    Depending on the search items, some information might overlap.

    Optional arguments:

    best: ``True``
        Return only the best match with the interfaces IP networks
        when the saerching pattern is a valid IP Address or Network.

    display: ``True``
        Display on the screen or return structured object? Default: `True` (return on the CLI).

    CLI Example:

    .. code-block:: bash

        $ sudo salt-run net.multi_find Ethernet1/49 xe-0/1/2

    Output Example:

    .. code-block:: text

        Pattern "Ethernet1/49" found in one of the following LLDP details

            -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
            |    Device    | Interface | Parent Interface | Remote Chassis ID | Remote Port Description | Remote Port ID |          Remote System Description          |   Remote System Name   |
            -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
            | edge01.oua04 |  xe-0/1/2 |       ae1        | DE:AD:BE:EF:DE:AD |       Ethernet1/49      |                | Cisco NX-OS(tm) n6000, Software (n6000-uk9) | edge07.oua04.dummy.net |
            -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

        Details for interface xe-0/1/2

            -----------------------------------------------------------------------------------------------------------------------
            |    Device    | Interface | Interface Description | IP Addresses | Enabled |  UP  |    MAC Address    | Speed [Mbps] |
            -----------------------------------------------------------------------------------------------------------------------
            | edge01.oua04 |  xe-0/1/2 |     ae1 sw01.oua04    |              |   True  | True | BE:EF:DE:AD:BE:EF |    10000     |
            -----------------------------------------------------------------------------------------------------------------------

        LLDP Neighbors for interface xe-0/1/2

            -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
            |    Device    | Interface | Parent Interface | Remote Chassis ID | Remote Port Description | Remote Port ID |          Remote System Description          |   Remote System Name   |
            -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
            | edge01.oua04 |  xe-0/1/2 |       ae1        | DE:AD:BE:EF:DE:AD |       Ethernet1/49      |                | Cisco NX-OS(tm) n6000, Software (n6000-uk9) | edge07.oua04.dummy.net |
            -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    """
    out = {}
    for pattern in set(patterns):
        search_result = find(
            pattern,
            best=kwargs.get("best", True),
            display=kwargs.get("display", _DEFAULT_DISPLAY),
        )
        out[pattern] = search_result
    if not kwargs.get("display", _DEFAULT_DISPLAY):
        return out



"""
If Salt's OS detection does not identify a different virtual service module, the minion will fall back to using this basic module, which simply wraps sysvinit scripts.
"""

import fnmatch
import os
import re

__func_alias__ = {"reload_": "reload"}

_GRAINMAP = {"Arch": "/etc/rc.d", "Arch ARM": "/etc/rc.d"}


def __virtual__():
    """
    Only work on systems which exclusively use sysvinit
    """
    # Disable on these platforms, specific service modules exist:
    disable = {
        "RedHat",
        "CentOS",
        "Amazon",
        "ScientificLinux",
        "CloudLinux",
        "Fedora",
        "Gentoo",
        "Ubuntu",
        "Debian",
        "Devuan",
        "ALT",
        "OEL",
        "Linaro",
        "elementary OS",
        "McAfee  OS Server",
        "Raspbian",
        "SUSE",
        "Slackware",
    }
    if __grains__.get("os") in disable:
        return (False, "Your OS is on the disabled list")
    # Disable on all non-Linux OSes as well
    if __grains__["kernel"] != "Linux":
        return (False, "Non Linux OSes are not supported")
    init_grain = __grains__.get("init")
    if init_grain not in (None, "sysvinit", "unknown"):
        return (False, "Minion is running {}".format(init_grain))
    elif __utils__["systemd.booted"](__context__):
        # Should have been caught by init grain check, but check just in case
        return (False, "Minion is running systemd")
    return "service"


def run(name, action):
    """
    Run the specified service with an action.

    .. versionadded:: 2015.8.1

    name
        Service name.

    action
        Action name (like start,  stop,  reload,  restart).

    CLI Example:

    .. code-block:: bash

        salt '*' service.run apache2 reload
        salt '*' service.run postgresql initdb
    """
    cmd = (
        os.path.join(_GRAINMAP.get(__grains__.get("os"), "/etc/init.d"), name)
        + " "
        + action
    )
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def start(name):
    """
    Start the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """
    return run(name, "start")


def stop(name):
    """
    Stop the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    return run(name, "stop")


def restart(name):
    """
    Restart the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """
    return run(name, "restart")


def status(name, sig=None):
    """
    Return the status for a service.
    If the name contains globbing, a dict mapping service name to PID or empty
    string is returned.

    .. versionchanged:: 2018.3.0
        The service name can now be a glob (e.g. ``salt*``)

    Args:
        name (str): The name of the service to check
        sig (str): Signature to use to find the service via ps

    Returns:
        string: PID if running, empty otherwise
        dict: Maps service name to PID if running, empty string otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name> [service signature]
    """
    if sig:
        return __salt__["status.pid"](sig)

    contains_globbing = bool(re.search(r"\*|\?|\[.+\]", name))
    if contains_globbing:
        services = fnmatch.filter(get_all(), name)
    else:
        services = [name]
    results = {}
    for service in services:
        results[service] = __salt__["status.pid"](service)
    if contains_globbing:
        return results
    return results[name]


def reload_(name):
    """
    Refreshes config files by calling service reload. Does not perform a full
    restart.

    CLI Example:

    .. code-block:: bash

        salt '*' service.reload <service name>
    """
    return run(name, "reload")


def available(name):
    """
    Returns ``True`` if the specified service is available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' service.available sshd
    """
    return name in get_all()


def missing(name):
    """
    The inverse of service.available.
    Returns ``True`` if the specified service is not available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' service.missing sshd
    """
    return name not in get_all()


def get_all():
    """
    Return a list of all available services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    if not os.path.isdir(_GRAINMAP.get(__grains__.get("os"), "/etc/init.d")):
        return []
    return sorted(os.listdir(_GRAINMAP.get(__grains__.get("os"), "/etc/init.d")))



"""
Authentication runner for creating, deleting, and managing eauth tokens.

.. versionadded:: 2016.11.0

"""


import os

import salt.auth
import salt.exceptions
import salt.netapi


def mk_token(**load):
    r"""
    Create an eauth token using provided credentials

    Non-root users may specify an expiration date -- if allowed via the
    :conf_master:`token_expire_user_override` setting -- by passing an
    additional ``token_expire`` param. This overrides the
    :conf_master:`token_expire` setting of the same name in the Master config
    and is how long a token should live in seconds.

    CLI Example:

    .. code-block:: shell

        salt-run auth.mk_token username=saltdev password=saltdev eauth=auto

        # Create a token valid for three years.
        salt-run auth.mk_token username=saltdev password=saltdev eauth=auto \
            token_expire=94670856

        # Calculate the number of seconds using expr.
        salt-run auth.mk_token username=saltdev password=saltdev eauth=auto \
            token_expire=$(expr \( 365 \* 24 \* 60 \* 60 \) \* 3)
    """
    # This will hang if the master daemon is not running.
    netapi = salt.netapi.NetapiClient(__opts__)
    if not netapi._is_master_running():
        raise salt.exceptions.SaltDaemonNotRunning("Salt Master must be running.")

    auth = salt.auth.Resolver(__opts__)
    return auth.mk_token(load)


def del_token(token):
    """
    Delete an eauth token by name

    CLI Example:

    .. code-block:: shell

        salt-run auth.del_token 6556760736e4077daa601baec2b67c24
    """
    token_path = os.path.join(__opts__["token_dir"], token)
    if os.path.exists(token_path):
        return os.remove(token_path) is None
    return False



import pytest

import salt.config
import salt.wheel


@pytest.fixture
def client_config(salt_master):
    config = salt.config.client_config(
        salt_master.config["conf_file"],
        defaults=salt_master.config.copy(),
    )
    return config


@pytest.fixture(scope="module")
def salt_auto_account(salt_auto_account_factory):
    with salt_auto_account_factory as account:
        yield account


@pytest.fixture
def auth_creds(salt_auto_account):
    return {
        "username": salt_auto_account.username,
        "password": salt_auto_account.password,
        "eauth": "auto",
    }
